import os
from pathlib import Path
from unittest import mock

import pytest
from eth_pydantic_types import HexBytes32
from eth_tester.exceptions import TransactionFailed  # type: ignore
from eth_typing import HexStr
from eth_utils import ValidationError, to_hex
from hexbytes import HexBytes
from requests import HTTPError
from web3.exceptions import ContractPanicError, ExtraDataLengthError, TimeExhausted

from ape import convert
from ape.api.providers import SubprocessProvider
from ape.exceptions import (
    APINotImplementedError,
    BlockNotFoundError,
    ContractLogicError,
    ProviderError,
    TransactionError,
    TransactionNotFoundError,
    UnknownSnapshotError,
)
from ape.types.events import LogFilter
from ape.utils.testing import DEFAULT_TEST_CHAIN_ID
from ape_ethereum.provider import (
    WEB3_PROVIDER_URI_ENV_VAR_NAME,
    EthereumNodeProvider,
    Web3Provider,
    _sanitize_web3_url,
)
from ape_ethereum.transactions import TransactionStatusEnum, TransactionType
from ape_test import LocalProvider


def test_uri(eth_tester_provider):
    assert not eth_tester_provider.http_uri
    assert not eth_tester_provider.ws_uri


@pytest.mark.parametrize("block_id", (0, "0", "0x0", HexStr("0x0")))
def test_get_block(eth_tester_provider, block_id, vyper_contract_instance, owner):
    block = eth_tester_provider.get_block(block_id)

    # Each parameter is the same as requesting the first block.
    assert block.number == 0
    assert block.base_fee == 1000000000
    assert block.gas_used == 0


def test_get_block_not_found(eth_tester_provider):
    latest_block = eth_tester_provider.get_block("latest")
    block_id = latest_block.number + 1000
    with pytest.raises(BlockNotFoundError, match=f"Block with ID '{block_id}' not found."):
        eth_tester_provider.get_block(block_id)


def test_get_block_transaction(vyper_contract_instance, owner, eth_tester_provider):
    # Ensure a transaction in latest block
    receipt = vyper_contract_instance.setNumber(900, sender=owner)
    block = eth_tester_provider.get_block(receipt.block_number)
    assert to_hex(block.transactions[-1].txn_hash) == receipt.txn_hash


def test_estimate_gas(vyper_contract_instance, eth_tester_provider, owner):
    txn = vyper_contract_instance.setNumber.as_transaction(900, sender=owner)
    estimate = eth_tester_provider.estimate_gas_cost(txn)
    assert estimate > 0


def test_estimate_gas_of_static_fee_txn(vyper_contract_instance, eth_tester_provider, owner):
    txn = vyper_contract_instance.setNumber.as_transaction(900, sender=owner, type=0)
    estimate = eth_tester_provider.estimate_gas_cost(txn)
    assert estimate > 0


def test_estimate_gas_with_max_value_from_block(
    mocker, eth_tester_provider, vyper_contract_instance
):
    mock_limit = mocker.patch(
        "ape.api.networks.NetworkAPI.gas_limit", new_callable=mock.PropertyMock
    )
    mock_limit.return_value = "max"
    txn = vyper_contract_instance.setNumber.as_transaction(900)
    gas_cost = eth_tester_provider.estimate_gas_cost(txn)
    latest_block = eth_tester_provider.get_block("latest")

    # NOTE: Gas is estimated if asked, regardless of network defaults.
    assert gas_cost < latest_block.gas_limit


def test_chain_id(eth_tester_provider):
    chain_id = eth_tester_provider.chain_id
    assert chain_id == DEFAULT_TEST_CHAIN_ID


def test_chain_id_is_cached(eth_tester_provider):
    _ = eth_tester_provider.chain_id

    # Unset `_web3` to show that it is not used in a second call to `chain_id`.
    web3 = eth_tester_provider._web3
    eth_tester_provider._web3 = None
    chain_id = eth_tester_provider.chain_id
    assert chain_id == DEFAULT_TEST_CHAIN_ID
    eth_tester_provider._web3 = web3  # Undo


def test_chain_id_from_ethereum_base_provider_is_cached(mock_web3, ethereum, eth_tester_provider):
    """
    Simulated chain ID from a plugin (using base-ethereum class) to ensure is
    also cached.
    """

    def make_request(rpc, arguments):
        if rpc == "eth_chainId":
            return {"result": 11155111}  # Sepolia

        return eth_tester_provider.make_request(rpc, arguments)

    mock_web3.provider.make_request.side_effect = make_request

    class PluginProvider(Web3Provider):
        def connect(self):
            return

        def disconnect(self):
            return

    provider = PluginProvider(name="sim", network=ethereum.sepolia)
    provider._web3 = mock_web3
    assert provider.chain_id == 11155111
    # Unset to web3 to prove it does not check it again (else it would fail).
    provider._web3 = None
    assert provider.chain_id == 11155111


def test_chain_id_when_disconnected(eth_tester_provider):
    eth_tester_provider.disconnect()
    try:
        actual = eth_tester_provider.chain_id
        expected = DEFAULT_TEST_CHAIN_ID
        assert actual == expected

    finally:
        eth_tester_provider.connect()


def test_chain_id_adhoc_http(networks):
    with networks.parse_network_choice("https://www.shibrpc.com") as bor:
        assert bor.chain_id == 109


def test_get_receipt_not_exists_with_timeout(eth_tester_provider):
    unknown_txn = "0x053cba5c12172654d894f66d5670bab6215517a94189a9ffc09bc40a589ec04d"
    expected = (
        f"Transaction '{unknown_txn}' not found. "
        rf"Error: Transaction '{unknown_txn}' "
        "is not in the chain after 0 seconds"
    )
    with pytest.raises(TransactionNotFoundError, match=expected):
        eth_tester_provider.get_receipt(unknown_txn, timeout=0)


def test_get_receipt_exists_with_timeout(eth_tester_provider, vyper_contract_instance, owner):
    receipt_from_invoke = vyper_contract_instance.setNumber(888, sender=owner)
    receipt_from_provider = eth_tester_provider.get_receipt(receipt_from_invoke.txn_hash, timeout=0)
    assert receipt_from_provider.txn_hash == receipt_from_invoke.txn_hash
    assert receipt_from_provider.receiver == vyper_contract_instance.address


def test_get_receipt_ignores_timeout_when_private(
    eth_tester_provider, mock_web3, vyper_contract_instance, owner
):
    receipt_from_invoke = vyper_contract_instance.setNumber(889, sender=owner)
    real_web3 = eth_tester_provider._web3

    mock_web3.eth.wait_for_transaction_receipt.side_effect = TimeExhausted
    eth_tester_provider._web3 = mock_web3
    try:
        receipt_from_provider = eth_tester_provider.get_receipt(
            receipt_from_invoke.txn_hash, timeout=5, private=True
        )

    finally:
        eth_tester_provider._web3 = real_web3

    assert receipt_from_provider.txn_hash == receipt_from_invoke.txn_hash
    assert not receipt_from_provider.confirmed


def test_get_receipt_passes_receipt_when_private(
    eth_tester_provider, mock_web3, vyper_contract_instance, owner
):
    receipt_from_invoke = vyper_contract_instance.setNumber(890, sender=owner)
    real_web3 = eth_tester_provider._web3

    mock_web3.eth.wait_for_transaction_receipt.side_effect = TimeExhausted
    eth_tester_provider._web3 = mock_web3
    try:
        receipt_from_provider = eth_tester_provider.get_receipt(
            receipt_from_invoke.txn_hash,
            timeout=5,
            private=True,
            transaction=receipt_from_invoke.transaction,
        )

    finally:
        eth_tester_provider._web3 = real_web3

    assert receipt_from_provider.txn_hash == receipt_from_invoke.txn_hash
    assert not receipt_from_provider.confirmed

    # Receiver comes from the transaction.
    assert receipt_from_provider.receiver == vyper_contract_instance.address


def test_get_contracts_logs_all_logs(chain, contract_instance, owner, eth_tester_provider):
    start_block = chain.blocks.height
    stop_block = start_block + 100
    log_filter = LogFilter(
        addresses=[contract_instance],
        events=contract_instance.contract_type.events,
        start_block=start_block,
        stop_block=stop_block,
    )
    logs_at_start = len([log for log in eth_tester_provider.get_contract_logs(log_filter)])
    contract_instance.fooAndBar(sender=owner)  # Create 2 logs
    logs_after_new_emit = [log for log in eth_tester_provider.get_contract_logs(log_filter)]
    assert len(logs_after_new_emit) == logs_at_start + 2


def test_get_contract_logs_single_log(chain, contract_instance, owner, eth_tester_provider):
    contract_instance.fooAndBar(sender=owner)  # Create logs
    block = chain.blocks.height
    log_filter = LogFilter.from_event(
        event=contract_instance.FooHappened,
        search_topics={"foo": 0},
        addresses=[contract_instance],
        start_block=block,
        stop_block=block,
    )
    logs = [log for log in eth_tester_provider.get_contract_logs(log_filter)]
    assert len(logs) == 1
    assert logs[0]["foo"] == 0
    assert logs[0].abi == contract_instance.FooHappened.abi

    # Show it looks up the ABI when not cached not anymore.
    logs[0]._abi = None
    assert logs[0].abi == contract_instance.FooHappened.abi

    # Ensure topics are expected.
    topics = logs[0].topics
    expected_topics = [
        "0x1a7c56fae0af54ebae73bc4699b9de9835e7bb86b050dff7e80695b633f17abd",
        "0x0000000000000000000000000000000000000000000000000000000000000000",
    ]
    assert topics == expected_topics


def test_get_contract_logs_single_log_query_multiple_values(
    chain, contract_instance, owner, eth_tester_provider
):
    contract_instance.fooAndBar(sender=owner)  # Create logs
    block = chain.blocks.height
    log_filter = LogFilter.from_event(
        event=contract_instance.FooHappened,
        search_topics={"foo": [0, 1]},
        addresses=[contract_instance],
        start_block=block,
        stop_block=block,
    )
    logs = [log for log in eth_tester_provider.get_contract_logs(log_filter)]
    assert len(logs) >= 1
    assert logs[-1]["foo"] == 0


def test_get_contract_logs_multiple_accounts_for_address(
    chain, contract_instance, owner, eth_tester_provider
):
    """
    Tests the condition when you pass in multiple AddressAPI objects
    during an address-topic search.
    """
    contract_instance.logAddressArray(sender=owner)  # Create logs
    block = chain.blocks.height
    log_filter = LogFilter.from_event(
        event=contract_instance.EventWithAddressArray,
        search_topics={"some_address": [owner, contract_instance]},
        addresses=[contract_instance, owner],
        start_block=block,
        stop_block=block,
    )
    logs = [log for log in eth_tester_provider.get_contract_logs(log_filter)]
    assert len(logs) >= 1
    assert logs[-1]["some_address"] == owner.address


def test_get_contract_logs_single_log_unmatched(
    chain, contract_instance, owner, eth_tester_provider
):
    unmatched_search = {"foo": 2}  # Foo is created with a value of 0
    contract_instance.fooAndBar(sender=owner)  # Create logs
    block = chain.blocks.height
    log_filter = LogFilter.from_event(
        event=contract_instance.FooHappened,
        search_topics=unmatched_search,
        addresses=[contract_instance],
        start_block=block,
        stop_block=block,
    )
    logs = [log for log in eth_tester_provider.get_contract_logs(log_filter)]
    assert len(logs) == 0


def test_supports_tracing(eth_tester_provider):
    assert not eth_tester_provider.supports_tracing


def test_get_balance(networks, accounts):
    balance = networks.provider.get_balance(accounts[0].address)
    assert type(balance) is int
    assert balance > 0


def test_set_timestamp(ethereum):
    # NOTE: Using a different eth-tester for multi-processing ease.
    with ethereum.local.use_provider(
        "test", provider_settings={"chain_id": 919191912828283}
    ) as provider:
        pending_at_start = provider.get_block("pending").timestamp
        new_ts = pending_at_start + 100
        expected = new_ts + 1  # Mining adds another second.
        provider.set_timestamp(new_ts)
        provider.mine()
        actual = provider.get_block("pending").timestamp
        assert actual == expected


def test_set_timestamp_to_same_time(eth_tester_provider):
    """
    Eth tester normally fails when setting the timestamp to the same time.
    However, in Ape, we treat it as a no-op and let it pass.
    """
    expected = eth_tester_provider.get_block("pending").timestamp
    eth_tester_provider.set_timestamp(expected)
    actual = eth_tester_provider.get_block("pending").timestamp
    assert actual == expected


def test_set_timestamp_handle_same_time_race_condition(mocker, eth_tester_provider):
    """
    Ensures that when we get an error saying the timestamps are the same,
    we ignore it and treat it as a noop. This handles the race condition
    when the block advances after ``set_timestamp`` has been called but before
    the operation completes.
    """

    def side_effect(*args, **kwargs):
        raise ValidationError(
            "timestamp must be strictly later than parent, "
            "but is 0 seconds before.\n"
            "- child  : 0\n"
            "- parent : 0."
        )

    mocker.patch.object(eth_tester_provider.evm_backend, "time_travel", side_effect=side_effect)
    eth_tester_provider.set_timestamp(123)


def test_get_virtual_machine_error_when_txn_failed_includes_base_error(
    eth_tester_provider,
):
    txn_failed = TransactionFailed()
    actual = eth_tester_provider.get_virtual_machine_error(txn_failed)
    assert actual.base_err == txn_failed


def test_get_virtual_machine_error_panic(eth_tester_provider, mocker):
    data = "0x4e487b710000000000000000000000000000000000000000000000000000000000000032"
    message = "Panic error 0x32: Array index is out of bounds."
    exception = ContractPanicError(data=data, message=message)
    enrich_spy = mocker.spy(eth_tester_provider.compiler_manager, "enrich_error")
    actual = eth_tester_provider.get_virtual_machine_error(exception)
    assert enrich_spy.call_count == 1
    enrich_spy.assert_called_once_with(actual)
    assert isinstance(actual, ContractLogicError)


def test_gas_price(eth_tester_provider):
    actual = eth_tester_provider.gas_price
    assert isinstance(actual, int)


def test_get_code(eth_tester_provider, vyper_contract_instance):
    address = vyper_contract_instance.address
    block_number = vyper_contract_instance.creation_metadata.block
    assert eth_tester_provider.get_code(address) == eth_tester_provider.get_code(
        address, block_id=block_number
    )


@pytest.mark.parametrize("tx_type", TransactionType)
def test_prepare_transaction_with_max_gas(tx_type, eth_tester_provider, ethereum, owner):
    tx = ethereum.create_transaction(type=tx_type.value, sender=owner.address)
    tx.gas_limit = None  # Undo set from validator
    assert tx.gas_limit is None, "Test setup failed - couldn't clear tx gas limit."

    actual = eth_tester_provider.prepare_transaction(tx)
    assert actual.gas_limit == eth_tester_provider.max_gas
    assert actual.max_fee is not None


def test_no_comma_in_rpc_url():
    test_url = "URI: http://127.0.0.1:8545,"
    sanitised_url = _sanitize_web3_url(test_url)

    assert "," not in sanitised_url


def test_send_transaction_when_no_error_and_receipt_fails(
    mocker,
    mock_web3,
    mock_transaction,
    eth_tester_provider,
    owner,
    vyper_contract_instance,
):
    start_web3 = eth_tester_provider._web3
    eth_tester_provider._web3 = mock_web3
    mock_eth_tester = mocker.MagicMock()
    original_tester = eth_tester_provider.tester
    eth_tester_provider.__dict__["tester"] = mock_eth_tester

    try:
        # NOTE: Value is meaningless.
        tx_hash = HexBytes32.__eth_pydantic_validate__(123**36)

        # Sending tx "works" meaning no vm error.
        mock_eth_tester.ethereum_tester.send_raw_transaction.return_value = tx_hash

        # Getting a receipt "works", but you get a failed one.
        receipt_data = {
            "failed": True,
            "blockNumber": 0,
            "txnHash": to_hex(tx_hash),
            "status": TransactionStatusEnum.FAILING.value,
            "sender": owner.address,
            "receiver": vyper_contract_instance.address,
            "input": b"",
            "gasUsed": 123,
            "gasLimit": 100,
        }
        mock_web3.eth.wait_for_transaction_receipt.return_value = receipt_data

        # Attempting to replay the tx does not produce any error.
        mock_web3.eth.call.return_value = HexBytes("")

        # Execute test.
        mock_transaction.serialize_transaction.return_value = HexBytes(123123123123)
        with pytest.raises(TransactionError):
            eth_tester_provider.send_transaction(mock_transaction)

    finally:
        eth_tester_provider._web3 = start_web3
        eth_tester_provider.__dict__["tester"] = original_tester


def test_network_choice(eth_tester_provider):
    actual = eth_tester_provider.network_choice
    expected = "ethereum:local:test"
    assert actual == expected


def test_network_choice_when_custom(eth_tester_provider):
    name = eth_tester_provider.network.name
    eth_tester_provider.network.name = "custom"
    try:
        # NOTE: Raises this error because EthTester does not support custom
        #   connections.
        with pytest.raises(
            ProviderError, match=".*Custom network provider missing `connection_str`.*"
        ):
            _ = eth_tester_provider.network_choice
    finally:
        eth_tester_provider.network.name = name


def test_make_request_not_exists(eth_tester_provider):
    with pytest.raises(
        APINotImplementedError,
        match="RPC method 'ape_thisDoesNotExist' is not implemented by this node instance.",
    ):
        eth_tester_provider.make_request("ape_thisDoesNotExist")


@pytest.mark.parametrize("msg", ("Method not found", "Method ape_thisDoesNotExist not found"))
def test_make_request_not_exists_dev_nodes(eth_tester_provider, mock_web3, msg):
    """
    Handle an issue found from Base-sepolia where not-implemented RPCs
    caused HTTPErrors.
    """
    real_web3 = eth_tester_provider._web3
    mock_web3.eth = real_web3.eth

    def custom_make_request(rpc, params):
        if rpc == "ape_thisDoesNotExist":
            return {"error": {"message": msg}}

        return real_web3.provider.make_request(rpc, params)

    mock_web3.provider.make_request.side_effect = custom_make_request

    eth_tester_provider._web3 = mock_web3
    try:
        with pytest.raises(
            APINotImplementedError,
            match="RPC method 'ape_thisDoesNotExist' is not implemented by this node instance.",
        ):
            eth_tester_provider.make_request("ape_thisDoesNotExist")
    finally:
        eth_tester_provider._web3 = real_web3


def test_make_request_handles_http_error_method_not_allowed(eth_tester_provider, mock_web3):
    """
    Simulate what *most* of the dev providers do, like hardhat, anvil, and ganache.
    """
    real_web3 = eth_tester_provider._web3
    mock_web3.eth = real_web3.eth

    def custom_make_request(rpc, params):
        if rpc == "ape_thisDoesNotExist":
            raise HTTPError("Client error: Method Not Allowed")

        return real_web3.provider.make_request(rpc, params)

    mock_web3.provider.make_request.side_effect = custom_make_request
    eth_tester_provider._web3 = mock_web3
    try:
        with pytest.raises(
            APINotImplementedError,
            match="RPC method 'ape_thisDoesNotExist' is not implemented by this node instance.",
        ):
            eth_tester_provider.make_request("ape_thisDoesNotExist")
    finally:
        eth_tester_provider._web3 = real_web3


def test_make_request_rate_limiting(mocker, ethereum, mock_web3):
    provider = EthereumNodeProvider(network=ethereum.local)
    provider._web3 = mock_web3

    class RateLimitTester:
        tries = 3
        _try = 0
        tries_made = 0

        def rate_limit_hook(self, rpc, params):
            self.tries_made += 1
            if self._try >= self.tries:
                self._try = 0
                return {"success": True}
            else:
                self._try += 1
                response = mocker.MagicMock()
                response.status_code = 429
                raise HTTPError(response=response)

    rate_limit_tester = RateLimitTester()
    mock_web3.provider.make_request.side_effect = rate_limit_tester.rate_limit_hook
    result = provider.make_request("ape_testRateLimiting", parameters=[])
    assert rate_limit_tester.tries_made == rate_limit_tester.tries + 1
    assert result == {"success": True}


def test_base_fee(eth_tester_provider):
    actual = eth_tester_provider.base_fee
    assert actual >= eth_tester_provider.get_block("pending").base_fee

    # NOTE: Mostly doing this to ensure we are calling the fee history
    #   RPC correctly. There was a bug where we were not.
    actual = eth_tester_provider._get_fee_history(0)
    assert "baseFeePerGas" in actual


def test_has_poa_history_block_data(mock_web3, ethereum, eth_tester_provider):
    class PluginProvider(EthereumNodeProvider):
        pass

    provider = PluginProvider(name="prov", network=ethereum.sepolia)
    provider._web3 = mock_web3

    key = "proofOfAuthorityData"
    mock_web3.eth.get_block.return_value = {key: 123}

    assert provider.has_poa_history


def test_has_poa_history_block_exception(mock_web3, ethereum, eth_tester_provider):
    class PluginProvider(EthereumNodeProvider):
        pass

    provider = PluginProvider(name="prov", network=ethereum.sepolia)
    provider._web3 = mock_web3
    mock_web3.eth.get_block.side_effect = ExtraDataLengthError
    assert provider.has_poa_history


def test_has_poa_history_checks_earliest_and_latest_block(mock_web3, ethereum, eth_tester_provider):
    class PluginProvider(EthereumNodeProvider):
        pass

    provider = PluginProvider(name="prov", network=ethereum.sepolia)
    provider._web3 = mock_web3

    def get_block_side_effect(block_id):
        if block_id == "earliest":
            return {"blockNumber": 0}
        elif block_id == "latest":
            return {"blockNumber": 1, "proofOfAuthorityData": 123}

    mock_web3.eth.get_block.side_effect = get_block_side_effect
    poa_detected = provider.has_poa_history
    assert mock_web3.eth.get_block.call_count == 2
    assert poa_detected


def test_has_poa_history_false(mock_web3, ethereum, eth_tester_provider):
    class PluginProvider(EthereumNodeProvider):
        pass

    provider = PluginProvider(name="prov", network=ethereum.sepolia)
    provider._web3 = mock_web3
    mock_web3.eth.get_block.return_value = {}
    assert not provider.has_poa_history


def test_create_access_list(eth_tester_provider, vyper_contract_instance, owner):
    tx = vyper_contract_instance.setNumber.as_transaction(123, sender=owner)
    with pytest.raises(APINotImplementedError):
        eth_tester_provider.create_access_list(tx)


def test_auto_mine(eth_tester_provider, owner):
    eth_tester_provider.auto_mine = False
    assert not eth_tester_provider.auto_mine

    block_before = eth_tester_provider.get_block("latest").number
    nonce_before = owner.nonce

    # NOTE: Before, this would wait until it timed out, because
    #  when auto mine is off, `ape-test` provider still waited
    #  for the receipt during send_transaction(). It should
    #  instead return early.
    tx = owner.transfer(owner, 123)
    assert not tx.confirmed
    assert tx.sender == owner.address
    assert tx.txn_hash is not None

    nonce_after_tx = owner.nonce
    block_after_tx = eth_tester_provider.get_block("latest").number
    assert nonce_before == nonce_after_tx, "Transaction should not have been mined."
    assert block_before == block_after_tx, "Block height should not have increased."

    eth_tester_provider.mine()
    block_after_mine = eth_tester_provider.get_block("latest").number
    assert block_after_mine > block_after_tx

    eth_tester_provider.auto_mine = True
    assert eth_tester_provider.auto_mine


def test_new_when_web3_provider_uri_set():
    """
    Tests against a confusing case where having an env var
    $WEB3_PROVIDER_URI caused web3.py to only ever use that RPC
    URL regardless of what was said in Ape's --network or config.
    Now, we raise an error to avoid having users think Ape's
    network system is broken.
    """
    os.environ[WEB3_PROVIDER_URI_ENV_VAR_NAME] = "TEST"
    expected = (
        rf"Ape does not support Web3\.py's environment variable "
        rf"\${WEB3_PROVIDER_URI_ENV_VAR_NAME}\. If you are using this environment "
        r"variable name incidentally, please use a different name\. If you are "
        r"trying to set the network in Web3\.py, please use Ape's `ape-config\.yaml` "
        r"or `--network` option instead\."
    )

    class MyProvider(Web3Provider):
        def connect(self):
            raise NotImplementedError()

        def disconnect(self):
            raise NotImplementedError()

    try:
        with pytest.raises(ProviderError, match=expected):
            _ = MyProvider(data_folder=None, name=None, network=None)

    finally:
        if WEB3_PROVIDER_URI_ENV_VAR_NAME in os.environ:
            del os.environ[WEB3_PROVIDER_URI_ENV_VAR_NAME]


def test_account_balance_state(project, eth_tester_provider, owner):
    amount = convert("100_000 ETH", int)

    with project.temp_config(test={"balance": amount}):
        # NOTE: Purposely using a different instance of the provider
        #   for better testing isolation.
        provider = LocalProvider(
            name="test",
            network=eth_tester_provider.network,
        )
        provider.connect()
        bal = provider.get_balance(owner.address)
        assert bal == amount


@pytest.mark.parametrize(
    "uri,key",
    [
        ("ws://example.com", "ws_uri"),
        ("wss://example.com", "ws_uri"),
        ("wss://example.com", "uri"),
    ],
)
def test_node_ws_uri(project, uri, key):
    node = project.network_manager.ethereum.sepolia.get_provider("node")
    assert node.ws_uri is None
    config = {"ethereum": {"sepolia": {key: uri}}}
    with project.temp_config(node=config):
        node = project.network_manager.ethereum.sepolia.get_provider("node")
        assert node.ws_uri == uri

        if key != "ws_uri":
            assert node.uri == uri
        # else: uri gets to set to random HTTP from default settings,
        # but we may want to change that behavior.
        # TODO: 0.9 investigate not using random if ws set.


@pytest.mark.parametrize("http_key", ("uri", "http_uri"))
def test_node_http_uri_with_ws_uri(project, http_key):
    http = "http://example.com"
    ws = "ws://example.com"
    # Showing `uri:` as an HTTP and `ws_uri`: as an additional ws.
    with project.temp_config(node={"ethereum": {"sepolia": {http_key: http, "ws_uri": ws}}}):
        node = project.network_manager.ethereum.sepolia.get_provider("node")
        assert node.uri == http
        assert node.http_uri == http
        assert node.ws_uri == ws


@pytest.mark.parametrize("key", ("uri", "ipc_path"))
def test_ipc_per_network(project, key):
    ipc = "path/to/example.ipc"
    with project.temp_config(node={"ethereum": {"sepolia": {key: ipc}}}):
        node = project.network_manager.ethereum.sepolia.get_provider("node")
        assert node.ipc_path == Path(ipc)
        if key == "uri":
            assert node.uri == ipc
        # else: uri ends up as a random HTTP URI from evmchains.
        # TODO: Do we want to change this in 0.9?


def test_snapshot(eth_tester_provider):
    snapshot = eth_tester_provider.snapshot()
    assert snapshot


def test_restore(eth_tester_provider, accounts):
    account = accounts[0]
    start_nonce = account.nonce
    snapshot = eth_tester_provider.snapshot()
    account.transfer(account, 0)
    eth_tester_provider.restore(snapshot)
    assert account.nonce == start_nonce


def test_restore_zero(eth_tester_provider):
    with pytest.raises(UnknownSnapshotError, match="Unknown snapshot ID '0'."):
        eth_tester_provider.restore(0)


def test_update_settings_invalidates_snapshots(eth_tester_provider, chain):
    snapshot = chain.snapshot()
    assert snapshot in chain._snapshots[eth_tester_provider.chain_id]
    eth_tester_provider.update_settings({})
    assert snapshot not in chain._snapshots[eth_tester_provider.chain_id]


def test_connect_uses_cached_chain_id(mocker, mock_web3, ethereum, eth_tester_provider):
    class PluginProvider(EthereumNodeProvider):
        pass

    web3_factory_patch = mocker.patch("ape_ethereum.provider._create_web3")
    web3_factory_patch.return_value = mock_web3

    class ChainIDTracker:
        call_count = 0

        def make_request(self, rpc, args):
            if rpc == "eth_chainId":
                self.call_count += 1
                return {"result": "0xaa36a7"}  # Sepolia

            return eth_tester_provider.make_request(rpc, args)

    chain_id_tracker = ChainIDTracker()
    mock_web3.provider.make_request.side_effect = chain_id_tracker.make_request

    provider = PluginProvider(name="node", network=ethereum.sepolia)
    provider.connect()
    assert chain_id_tracker.call_count == 1
    provider.disconnect()
    provider.connect()
    # It is still cached from the previous connection.
    assert chain_id_tracker.call_count == 1


class TestSubprocessProvider:
    FAKE_PID = 12345678901234567890

    @pytest.fixture(autouse=True)
    def mock_process(self, mocker):
        mock_process = mocker.MagicMock()
        mock_process.pid = self.FAKE_PID
        return mock_process

    @pytest.fixture(autouse=True)
    def popen_patch(self, mocker, mock_process):
        # Prevent actually creating new processes.
        patch = mocker.patch("ape.api.providers.popen")
        patch.return_value = mock_process
        return patch

    @pytest.fixture(autouse=True)
    def spawn_patch(self, mocker):
        # Prevent spawning process monitoring threads.
        return mocker.patch("ape.api.providers.spawn")

    @pytest.fixture
    def subprocess_provider(self, popen_patch, eth_tester_provider):
        class MockSubprocessProvider(SubprocessProvider):
            @property
            def is_connected(self):
                # Once Popen is called once, we are "connected"
                return popen_patch.call_count > 0

            def build_command(self) -> list[str]:
                return ["apemockprocess"]

        # Hack to allow abstract methods anyway.
        MockSubprocessProvider.__abstractmethods__ = set()  # type: ignore

        return MockSubprocessProvider(name="apemockprocess", network=eth_tester_provider.network)  # type: ignore

    def test_start(self, subprocess_provider):
        assert not subprocess_provider.is_connected
        subprocess_provider.start()
        assert subprocess_provider.is_connected

        # Show it gets tracked in network manager's managed nodes.
        assert self.FAKE_PID in subprocess_provider.network_manager.running_nodes

    def test_start_allow_start_false(self, subprocess_provider):
        subprocess_provider.allow_start = False
        expected = r"Process not started and cannot connect to existing process\."
        with pytest.raises(ProviderError, match=expected):
            subprocess_provider.start()



"""Base non-pydantic validator utils"""

import re
from warnings import warn

from eth_utils import is_hex

from ape.exceptions import AccountsError, AliasAlreadyInUseError
from ape.utils.basemodel import ManagerAccessMixin

MIN_PASSPHRASE_LENGTH = 6


def _has_num(val: str):
    return re.search(r"\d{1}", val) is not None


def _has_special(val: str):
    return re.search(r"[\!@#$%\^&\*\(\) ]{1}", val) is not None


def _validate_account_alias(alias: str) -> str:
    """Validate an account alias"""
    if alias in ManagerAccessMixin.account_manager.aliases:
        raise AliasAlreadyInUseError(alias)
    elif not isinstance(alias, str):
        raise AccountsError(f"Alias must be a str, not '{type(alias)}'.")
    elif is_hex(alias) and len(alias) >= 42:
        # Prevents private keys from accidentally being stored in plaintext
        # Ref: https://github.com/ApeWorX/ape/issues/1525
        raise AccountsError("Longer aliases cannot be hex strings.")

    return alias


def _validate_account_passphrase(passphrase: str) -> str:
    """Make sure given passphrase is valid for account encryption"""
    if not passphrase or not isinstance(passphrase, str):
        raise AccountsError("Account file encryption passphrase must be provided.")

    if len(passphrase) < MIN_PASSPHRASE_LENGTH:
        warn("Passphrase length is extremely short. Consider using something longer.")

    if not (_has_num(passphrase) or _has_special(passphrase)):
        warn("Passphrase complexity is simple. Consider using numbers and special characters.")

    return passphrase


__all__ = ["_validate_account_alias", "_validate_account_passphrase"]



import copy
import shutil
from pathlib import Path

import click
import pytest
from click import BadParameter

from ape.cli import (
    AccountAliasPromptChoice,
    ConnectedProviderCommand,
    NetworkChoice,
    PromptChoice,
    account_option,
    config_override_option,
    contract_file_paths_argument,
    existing_alias_argument,
    network_option,
    non_existing_alias_argument,
    project_option,
    select_account,
    verbosity_option,
)
from ape.cli.choices import _NONE_NETWORK, _get_networks_sequence_from_cache
from ape.cli.commands import get_param_from_ctx, parse_network
from ape.exceptions import AccountsError
from ape.logging import LogLevel, logger
from tests.conftest import geth_process_test, skip_if_plugin_installed

OUTPUT_FORMAT = "__TEST__{0}:{1}:{2}_"
OTHER_OPTION_VALUE = "TEST_OTHER_OPTION"
other_option = click.option("--other", default=OTHER_OPTION_VALUE)


@pytest.fixture
def keyfile_swap_paths(config):
    return config.DATA_FOLDER / "accounts", config.DATA_FOLDER.parent / "temp_accounts"


@pytest.fixture
def one_keyfile_account(keyfile_swap_paths, keyfile_account, project):
    src_path, dest_path = keyfile_swap_paths
    existing_keyfiles = [x for x in src_path.iterdir() if x.is_file()]
    test_data = {"test": {"number_of_accounts": 0}}
    if existing_keyfiles == [keyfile_account.keyfile_path]:
        # Already only has the 1 account
        with project.temp_config(**test_data):
            yield keyfile_account

    else:
        if dest_path.is_file():
            dest_path.unlink()
        elif dest_path.is_dir():
            shutil.rmtree(dest_path)

        dest_path.mkdir()
        for keyfile in [x for x in existing_keyfiles if x != keyfile_account.keyfile_path]:
            shutil.copy(keyfile, dest_path / keyfile.name)
            keyfile.unlink()

        with project.temp_config(**test_data):
            yield keyfile_account

        for file in dest_path.iterdir():
            shutil.copy(file, src_path / file.name)


@pytest.fixture
def network_cmd():
    @click.command()
    @network_option()
    def cmd(ecosystem, network, provider):
        output = OUTPUT_FORMAT.format(ecosystem.name, network.name, provider.name)
        click.echo(output)

    return cmd


@pytest.fixture
def contracts_paths_cmd():
    expected = "EXPECTED {}"

    @click.command()
    @contract_file_paths_argument()
    @project_option()
    def cmd(file_paths, project):
        _ = project  # used in `contract_file_paths_argument`
        output = ", ".join(x.name for x in sorted(file_paths))
        click.echo(expected.format(output))

    return cmd


def _setup_temp_acct_number_change(accounts, num_accounts: int):
    if "containers" in accounts.__dict__:
        del accounts.__dict__["containers"]

    installed_account_types = {str(type(a)) for a in accounts}
    if installed_account_types:
        accounts_str = ", ".join(installed_account_types)
        pytest.fail(f"Unable to side-step install of account type(s): {accounts_str}")

    return {"test": {"number_of_accounts": num_accounts}}


def _teardown_numb_acct_change(accounts):
    if "containers" in accounts.__dict__:
        del accounts.__dict__["containers"]


@pytest.fixture
def no_accounts(account_manager, empty_data_folder, project):
    data = _setup_temp_acct_number_change(account_manager, 0)
    with project.temp_config(**data):
        yield

    _teardown_numb_acct_change(account_manager)


@pytest.fixture
def one_account(account_manager, empty_data_folder, project):
    data = _setup_temp_acct_number_change(account_manager, 1)
    with project.temp_config(**data):
        yield account_manager.test_accounts[0]

    _teardown_numb_acct_change(account_manager)


def get_expected_account_str(acct):
    return f"__expected_output__: {acct.address}"


def test_select_account_no_accounts_found(no_accounts):
    with pytest.raises(AccountsError, match="No accounts found."):
        assert not select_account()


def test_select_account_one_account(runner, one_account):
    # No input needed when only one account
    account = select_account()
    assert account == one_account


def test_select_account_multiple_accounts_requires_input(
    runner, keyfile_account, second_keyfile_account
):
    with runner.isolation(input="0\n"):
        account = select_account()

    assert account == keyfile_account


def test_select_account_custom_prompt(runner, keyfile_account, second_keyfile_account):
    prompt = "THIS_IS_A_CUSTOM_PROMPT"
    with runner.isolation(input="0\n") as out_streams:
        select_account(prompt)
        output = out_streams[0].getvalue().decode()

    assert prompt in output


def test_select_account_specify_type(runner, one_keyfile_account):
    with runner.isolation():
        account = select_account(key=type(one_keyfile_account))

    assert account == one_keyfile_account


def test_select_account_unknown_type(runner, keyfile_account):
    with pytest.raises(AccountsError) as err:
        select_account(key=str)  # type: ignore

    assert "Cannot return accounts with type '<class 'str'>'" in str(err.value)


def test_select_account_with_account_list(runner, keyfile_account, second_keyfile_account):
    account = select_account(key=[keyfile_account])
    assert account == keyfile_account

    account = select_account(key=[second_keyfile_account])
    assert account == second_keyfile_account

    with runner.isolation(input="1\n"):
        account = select_account(key=[keyfile_account, second_keyfile_account])
        assert account == second_keyfile_account


def test_network_option_default(runner, network_cmd):
    result = runner.invoke(network_cmd, catch_exceptions=False)
    assert result.exit_code == 0, result.output
    assert OUTPUT_FORMAT.format("ethereum", "local", "test") in result.output


def test_network_option_specified(runner, network_cmd):
    network_part = ("--network", "ethereum:local:test")
    result = runner.invoke(network_cmd, network_part)
    assert result.exit_code == 0, result.output
    assert OUTPUT_FORMAT.format("ethereum", "local", "test") in result.output


def test_network_option_unknown(runner, network_cmd):
    network_part = ("--network", "UNKNOWN")
    result = runner.invoke(network_cmd, network_part)
    assert result.exit_code != 0


def test_network_option_with_other_option(runner):
    """
    To prove can use the `@network_option` with other options
    in the same command (was issue during production where could not!).
    """

    # Scenario: Using network_option but not using the value in the command callback.
    #  (Potentially handling independently).
    @click.command()
    @network_option()
    @other_option
    def solo_option(other):
        click.echo(other)

    # Scenario: Using the network option with another option.
    # This use-case is way more common than the one above.
    @click.command()
    @network_option()
    @other_option
    def with_net(network, other):
        click.echo(network.name)
        click.echo(other)

    def run(cmd, fail_msg=None):
        res = runner.invoke(cmd, (), catch_exceptions=False)
        fail_msg = f"{fail_msg}\n{res.output}" if fail_msg else res.output
        assert res.exit_code == 0, fail_msg
        assert OTHER_OPTION_VALUE in res.output, fail_msg
        return res

    run(solo_option, fail_msg="Failed when used without network kwargs")
    result = run(with_net, fail_msg="Failed when used with network kwargs")
    assert "local" in result.output


@pytest.mark.parametrize(
    "network_input",
    (
        "ethereum:custom:https://127.0.0.1:4545",
        "ethereum:custom:https://127.0.0.1",
        "ethereum:custom:http://127.0.0.1:4545",
        "ethereum:custom:http://127.0.0.1",
        "ethereum:custom:http://foo.bar",
        "ethereum:custom:https://foo.bar:8000",
        ":custom:https://foo.bar:8000",
        "::https://foo.bar:8000",
        "https://foo.bar:8000",
    ),
)
def test_network_option_custom_uri(runner, network_cmd, network_input):
    network_part = ("--network", network_input)
    result = runner.invoke(network_cmd, network_part)
    assert result.exit_code == 0, result.output
    assert "custom" in result.output


def test_network_option_existing_network_with_custom_uri(runner, network_cmd):
    network_part = ("--network", "ethereum:sepolia:https://foo.bar:8000")
    result = runner.invoke(network_cmd, network_part)
    assert result.exit_code == 0, result.output
    assert "sepolia" in result.output


def test_network_option_make_required(runner):
    @click.command()
    @network_option(required=True)
    def cmd(network):
        click.echo(OUTPUT_FORMAT.format(network))

    result = runner.invoke(cmd, ())
    assert result.exit_code == 2
    assert "Error: Missing option '--network'." in result.output


def test_network_option_default_none(runner):
    @click.command()
    @network_option(default=None)
    def cmd(network):
        click.echo(f"Value is '{network}'")

    result = runner.invoke(cmd)
    assert "Value is 'None'" in result.output


def test_network_option_specified_none(runner):
    @click.command()
    @network_option()
    def cmd(network):
        click.echo(f"Value is '{network}'")

    result = runner.invoke(cmd, ("--network", "None"))
    assert "Value is 'None'" in result.output


@pytest.mark.parametrize("network_name", ("apenet", "apenet1"))
def test_network_option_specify_custom_network(
    runner, project, custom_networks_config_dict, network_name
):
    with project.temp_config(**custom_networks_config_dict):
        # NOTE: Also testing network filter with a custom network
        #  But this is also required to work around LRU cache
        #  giving us the wrong networks because click is running
        #  the tester in-process after re-configuring networks,
        #  which shouldn't happen IRL.

        @click.command()
        @network_option(network=network_name)
        def cmd(network):
            click.echo(f"Value is '{getattr(network, 'name', network)}'")

        result = runner.invoke(
            cmd, ("--network", f"ethereum:{network_name}:node"), catch_exceptions=False
        )
        assert result.exit_code == 0
        assert f"Value is '{network_name}'" in result.output

        # Fails because node is not a fork provider.
        result = runner.invoke(cmd, ("--network", f"ethereum:{network_name}-fork:node"))
        assert result.exit_code != 0
        assert f"No provider named 'node' in network '{network_name}-fork'" in result.output


def test_account_option(runner, keyfile_account):
    @click.command()
    @account_option()
    def cmd(account):
        _expected = get_expected_account_str(account)
        click.echo(_expected)

    expected = get_expected_account_str(keyfile_account)
    result = runner.invoke(cmd, ("--account", keyfile_account.alias))
    assert expected in result.output


def test_account_option_uses_single_account_as_default(runner, one_account):
    """
    When there is only 1 test account, that is the default
    when no option is given.
    """

    @click.command()
    @account_option(account_type=[one_account])
    def cmd(account):
        _expected = get_expected_account_str(account)
        click.echo(_expected)

    expected = get_expected_account_str(one_account)
    result = runner.invoke(cmd, ())
    assert expected in result.output


def test_account_prompts_when_more_than_one_keyfile_account(
    runner, keyfile_account, second_keyfile_account
):
    @click.command()
    @account_option()
    def cmd(account):
        _expected = get_expected_account_str(account)
        click.echo(_expected)

    expected = get_expected_account_str(keyfile_account)

    # Requires user input.
    result = runner.invoke(cmd, (), input="0\n")

    assert expected in result.output


@pytest.mark.parametrize("test_key", ("test", "TEST"))
def test_account_option_can_use_test_account(runner, accounts, test_key):
    index = 7
    test_account = accounts[index]

    @click.command()
    @account_option()
    def cmd(account):
        _expected = get_expected_account_str(account)
        click.echo(_expected)

    expected = get_expected_account_str(test_account)
    result = runner.invoke(cmd, ("--account", f"{test_key}::{index}"))
    assert expected in result.output


def test_account_option_alias_not_found(runner, keyfile_account):
    @click.command()
    @account_option()
    def cmd(account):
        pass

    result = runner.invoke(cmd, ("--account", "THIS ALAS IS NOT FOUND"))
    expected = (
        "Invalid value for '--account': Account with alias 'THIS ALAS IS NOT FOUND' not found"
    )
    assert expected in result.output


@pytest.mark.parametrize("opt", (0, "foo"))
def test_prompt_choice(runner, opt):
    """
    This demonstrates how to use ``PromptChoice``,
    as it is a little confusing, requiring a callback.
    """

    def choice_callback(ctx, param, value):
        return param.type.select()

    choice = PromptChoice(["foo", "bar"])
    assert hasattr(choice, "name")
    choice = PromptChoice(["foo", "bar"], name="choice")
    assert choice.name == "choice"

    @click.command()
    @click.option(
        "--choice",
        type=choice,
        callback=choice_callback,
    )
    def cmd(choice):
        click.echo(f"__expected_{choice}")

    result = runner.invoke(cmd, [], input=f"{opt}\n")
    assert "Select one of the following:" in result.output
    assert "__expected_foo" in result.output


@pytest.mark.parametrize("name", ("-v", "--verbosity"))
def test_verbosity_option(runner, name):
    logger._did_parse_sys_argv = False  # Force re-parse

    @click.command()
    @verbosity_option()
    def cmd():
        click.echo(f"__expected_{logger.level}")

    result = runner.invoke(cmd, (name, "debug"))
    assert "__expected_10" in result.output


@pytest.mark.parametrize(
    "level", (LogLevel.WARNING, LogLevel.WARNING.name, LogLevel.WARNING.value, "LogLevel.WARNING")
)
def test_verbosity_option_change_default(runner, level):
    @click.command()
    @verbosity_option(default=level)
    def cmd():
        pass

    verbosity_parameter = cmd.params[0]
    assert verbosity_parameter.default == level


def test_verbosity_option_uses_logger_level_as_default(runner):
    with logger.at_level(LogLevel.DEBUG):

        @click.command()
        @verbosity_option(default=None)
        def cmd():
            click.echo(f"LogLevel={logger.level}")

        result = runner.invoke(cmd)
        assert "LogLevel=10" in result.output


def test_account_prompt_name():
    """
    It is very important for this class to have the `name` attribute,
    even though it is not used. That is because some click internals
    expect this property to exist, and we skip the super() constructor.
    """
    option = AccountAliasPromptChoice()
    assert option.name == "account"
    option = AccountAliasPromptChoice(name="account_z")
    assert option.name == "account_z"


def test_contract_file_paths_argument_given_source_id(
    project_with_source_files_contract, runner, contracts_paths_cmd
):
    pm = project_with_source_files_contract
    src_id = next(x for x in pm.sources if Path(x).suffix == ".json")
    arguments = (src_id, "--project", f"{pm.path}")
    result = runner.invoke(contracts_paths_cmd, arguments)

    assert f"EXPECTED {src_id.split('/')[-1]}" in result.output


def test_contract_file_paths_argument_given_name(
    project_with_source_files_contract, runner, contracts_paths_cmd
):
    pm = project_with_source_files_contract
    src_stem = next(x for x in pm.sources if Path(x).suffix == ".json").split(".")[0]
    arguments = (src_stem, "--project", f"{pm.path}")
    result = runner.invoke(contracts_paths_cmd, arguments)

    assert f"EXPECTED {src_stem.split('/')[-1]}" in result.output


def test_contract_file_paths_argument_given_contracts_folder(
    project_with_contract, runner, contracts_paths_cmd
):
    pm = project_with_contract
    contracts_dirname = pm.contracts_folder.as_posix()
    arguments = (contracts_dirname, "--project", f"{pm.path}")
    result = runner.invoke(contracts_paths_cmd, arguments)
    all_paths = ", ".join(x.name for x in sorted(pm.sources.paths))

    assert f"EXPECTED {all_paths}" in result.output


def test_contract_file_paths_argument_given_contracts_folder_name(
    project_with_contract, runner, contracts_paths_cmd
):
    pm = project_with_contract
    arguments = ("contracts", "--project", f"{pm.path}")
    result = runner.invoke(contracts_paths_cmd, arguments)
    all_paths = ", ".join(x.name for x in sorted(pm.sources.paths))

    assert f"EXPECTED {all_paths}" in result.output


def test_contract_file_paths_argument_handles_exclude(
    project_with_contract, runner, contracts_paths_cmd
):
    pm = project_with_contract
    cfg = pm.config.get_config("compile")
    failmsg = "Setup failed - missing exclude config (set in ape-config.yaml)."
    assert "*Excl*" in cfg.exclude, failmsg

    # make a .cache file to show it is ignored.
    cache_file = project_with_contract.contracts_folder / ".cache" / "thing.json"
    cache_file.parent.mkdir(parents=True, exist_ok=True)
    cache_file.write_text("FAILS IF LOADED", encoding="utf8")

    result = runner.invoke(contracts_paths_cmd, "contracts")
    assert "Exclude.json" not in result.output
    assert "ExcludeNested.json" not in result.output
    # Ensure .cache always ignored!
    assert ".cache" not in result.output


@pytest.mark.parametrize("name", ("contracts/subdir", "subdir"))
def test_contract_file_paths_argument_given_subdir_relative_to_path(
    project_with_contract, runner, contracts_paths_cmd, name
):
    pm = project_with_contract
    arguments = (name, "--project", f"{pm.path}")
    result = runner.invoke(contracts_paths_cmd, arguments)
    paths = sorted(pm.sources.paths)

    all_paths = ", ".join(x.name for x in paths if x.parent.name == "subdir")
    assert f"EXPECTED {all_paths}" in result.output


@skip_if_plugin_installed("vyper")
def test_contract_file_paths_argument_missing_vyper(
    project_with_source_files_contract, runner, contracts_paths_cmd, ape_caplog
):
    name = "VyperContract"
    pm = project_with_source_files_contract
    arguments = (name, "--project", f"{pm.path}")
    with ape_caplog.at_level(LogLevel.INFO):
        result = runner.invoke(contracts_paths_cmd, arguments)

    expected = (
        "Missing compilers for the following file types: '.vy'. "
        "Possibly, a compiler plugin is not installed or is installed "
        "but not loading correctly. Is 'ape-vyper' installed?"
    )
    assert expected in result.output


@skip_if_plugin_installed("solidity")
def test_contract_file_paths_argument_missing_solidity(
    project_with_source_files_contract, runner, contracts_paths_cmd
):
    name = "SolidityContract"
    pm = project_with_source_files_contract
    with pm.isolate_in_tempdir() as tmp_project:
        arguments = (name, "--project", f"{tmp_project.path}")
        with logger.at_level(LogLevel.WARNING):
            result = runner.invoke(contracts_paths_cmd, arguments)

    expected = (
        "Missing compilers for the following file types: '.sol'. "
        "Possibly, a compiler plugin is not installed or is installed "
        "but not loading correctly. Is 'ape-solidity' installed?"
    )
    assert expected in result.output


def test_contract_file_paths_argument_contract_does_not_exist(
    project_with_source_files_contract, runner, contracts_paths_cmd
):
    name = "MadeUp"
    pm = project_with_source_files_contract
    with pm.isolate_in_tempdir() as tmp_project:
        arguments = (name, "--project", f"{tmp_project.path}")
        result = runner.invoke(contracts_paths_cmd, arguments)

    expected = f"Source file '{name}' not found."
    assert expected in result.output


def test_contract_file_paths_argument_given_directory_and_file(
    project_with_contract, runner, contracts_paths_cmd
):
    """
    Tests against a bug where if given a directory AND a file together,
    only the directory resolved and the file was lost.
    """
    pm = project_with_contract
    src_stem = next(x for x in pm.sources if Path(x).suffix == ".json").split(".")[0]
    arguments = ("subdir", src_stem, "--project", f"{pm.path}")
    result = runner.invoke(contracts_paths_cmd, arguments)
    paths = sorted(pm.sources.paths)

    all_paths = ", ".join(x.name for x in paths if x.parent.name == "subdir")
    assert f"{all_paths}" in result.output
    assert f"{src_stem.split('/')[-1]}" in result.output


def test_existing_alias_option(runner):
    @click.command()
    @existing_alias_argument()
    def cmd(alias):
        click.echo(alias)

    result = runner.invoke(cmd, "TEST::0")
    assert "TEST::0" in result.output


def test_existing_alias_option_custom_callback(runner):
    magic_value = "THIS IS A TEST"

    def custom_callback(*args, **kwargs):
        return magic_value

    @click.command()
    @existing_alias_argument(callback=custom_callback)
    def cmd(alias):
        click.echo(alias)

    result = runner.invoke(cmd, "TEST::0")
    assert magic_value in result.output


def test_non_existing_alias_option(runner):
    @click.command()
    @non_existing_alias_argument()
    def cmd(alias):
        click.echo(alias)

    result = runner.invoke(cmd, "non-exists")
    assert "non-exists" in result.output


def test_non_existing_alias_option_custom_callback(runner):
    magic_value = "THIS IS A TEST"

    def custom_callback(*args, **kwargs):
        return magic_value

    @click.command()
    @non_existing_alias_argument(callback=custom_callback)
    def cmd(alias):
        click.echo(alias)

    result = runner.invoke(cmd, "non-exists")
    assert magic_value in result.output


def test_connected_provider_command_no_args_or_network_specified(runner):
    @click.command(cls=ConnectedProviderCommand)
    def cmd():
        from ape import chain

        click.echo(chain.provider.is_connected)

    result = runner.invoke(cmd)
    assert result.exit_code == 0
    assert "True" in result.output, result.output


def test_connected_provider_command_invalid_value(runner):
    @click.command(cls=ConnectedProviderCommand)
    def cmd():
        pass

    result = runner.invoke(cmd, ("--network", "OOGA_BOOGA"), catch_exceptions=False)
    assert result.exit_code != 0
    assert "Invalid value for '--network'" in result.output


def test_connected_provider_command_use_provider(runner):
    @click.command(cls=ConnectedProviderCommand)
    def cmd(provider):
        click.echo(provider.is_connected)

    result = runner.invoke(cmd)
    assert result.exit_code == 0
    assert "True" in result.output, result.output


def test_connected_provider_command_use_ecosystem_network_and_provider(runner):
    @click.command(cls=ConnectedProviderCommand)
    def cmd(ecosystem, network, provider):
        click.echo(f"{ecosystem.name}:{network.name}:{provider.name}")

    result = runner.invoke(cmd)
    assert result.exit_code == 0
    assert "ethereum:local:test" in result.output, result.output


def test_connected_provider_command_use_ecosystem_network_and_provider_with_network_specified(
    runner,
):
    @click.command(cls=ConnectedProviderCommand)
    def cmd(ecosystem, network, provider):
        click.echo(f"{ecosystem.name}:{network.name}:{provider.name}")

    result = runner.invoke(cmd, ["--network", "ethereum:local:test"])
    assert result.exit_code == 0
    assert "ethereum:local:test" in result.output, result.output


def test_connected_provider_command_use_custom_options(runner):
    """
    Ensure custom options work when using `ConnectedProviderCommand`.
    (There was an issue during development where we could not).
    """

    # Scenario: Custom option and using network object.
    @click.command(cls=ConnectedProviderCommand)
    @other_option
    def use_net(network, other):
        click.echo(network.name)
        click.echo(other)

    # Scenario: Only using custom option.
    @click.command(cls=ConnectedProviderCommand)
    @other_option
    def solo_other(other):
        click.echo(other)

    @click.command(cls=ConnectedProviderCommand)
    @click.argument("other_arg")
    @other_option
    def with_arg(other_arg, other, provider):
        click.echo(other)
        click.echo(provider.name)
        click.echo(other_arg)

    spec = ("--network", "ethereum:local:test")

    def run(cmd, extra_args=None):
        arguments = [*spec, *(extra_args or [])]
        res = runner.invoke(cmd, arguments, catch_exceptions=False)
        assert res.exit_code == 0, res.output
        assert OTHER_OPTION_VALUE in res.output
        return res

    result = run(use_net)
    assert "local" in result.output, result.output  # Echos network object

    result = run(solo_other)
    assert "local" not in result.output, result.output

    argument = "_extra_"
    result = run(with_arg, extra_args=[argument])
    assert "test" in result.output
    assert argument in result.output


@geth_process_test
def test_connected_provider_command_with_network_option(runner, geth_provider):
    _ = geth_provider  # Ensure already running, to avoid clashing later on.

    @click.command(cls=ConnectedProviderCommand)
    @network_option()
    def cmd(provider):
        click.echo(provider.name)

    # NOTE: Must use a network that is not the default.
    spec = ("--network", "ethereum:local:node")
    res = runner.invoke(cmd, spec, catch_exceptions=False)
    assert res.exit_code == 0, res.output
    assert "node" in res.output


@geth_process_test
def test_connected_provider_command_with_network_option_and_cls_types_false(runner, geth_provider):
    _ = geth_provider  # Ensure already running, to avoid clashing later on.

    @click.command(cls=ConnectedProviderCommand, use_cls_types=False)
    @network_option()
    def cmd(network):
        assert isinstance(network, str)
        assert network.startswith("ethereum:local")

    # NOTE: Must use a network that is not the default.
    spec = ("--network", "ethereum:local:node")
    res = runner.invoke(cmd, spec, catch_exceptions=False)
    assert res.exit_code == 0, res.output


def test_connected_provider_command_none_network(runner):
    @click.command(cls=ConnectedProviderCommand)
    def cmd(network, provider):
        click.echo(network)
        click.echo(provider)

    spec = ("--network", "None")
    res = runner.invoke(cmd, spec, catch_exceptions=False)
    assert res.exit_code == 0, res.output


def test_get_param_from_ctx(mocker):
    mock_ctx = mocker.MagicMock()
    mock_ctx.params = {"foo": "bar"}
    mock_ctx.parent = mocker.MagicMock()
    mock_ctx.parent.params = {"interactive": True}
    actual = get_param_from_ctx(mock_ctx, "interactive")
    assert actual is True


def test_parse_network_when_interactive_and_no_param(mocker):
    ctx = mocker.MagicMock()
    ctx.params = {"interactive": True}
    ctx.parent = None
    network_ctx = parse_network(ctx)
    assert network_ctx is not None
    assert network_ctx.provider.name == "test"
    assert network_ctx._disconnect_on_exit is False  # Because of interactive: True


def test_parse_network_when_interactive_and_str_param(mocker):
    ctx = mocker.MagicMock()
    ctx.params = {"interactive": True, "network": "ethereum:local:test"}
    network_ctx = parse_network(ctx)
    assert network_ctx is not None
    assert network_ctx.provider.name == "test"
    assert network_ctx._disconnect_on_exit is False  # Because of interactive: True


def test_parse_network_when_interactive_and_class_param(mocker, eth_tester_provider):
    ctx = mocker.MagicMock()
    ctx.params = {"interactive": True, "network": eth_tester_provider}
    network_ctx = parse_network(ctx)
    assert network_ctx is not None
    assert network_ctx.provider.name == "test"
    assert network_ctx._disconnect_on_exit is False  # Because of interactive: True


def test_parse_network_when_explicit_none(mocker):
    ctx = mocker.MagicMock()
    ctx.params = {"network": _NONE_NETWORK}
    network_ctx = parse_network(ctx)
    assert network_ctx is None


class TestNetworkChoice:
    @pytest.fixture
    def network_choice(self):
        return NetworkChoice()

    def test_test(self, network_choice):
        actual = network_choice.convert("ethereum:local:test", None, None)
        assert actual.name == "test"
        assert actual.network.name == "local"

    @pytest.mark.parametrize("prefix", ("", "ethereum:custom:"))
    def test_adhoc(self, network_choice, prefix):
        uri = "https://example.com"
        actual = network_choice.convert(f"{prefix}{uri}", None, None)
        assert actual.uri == uri
        assert actual.network.name == "custom"

    def test_custom_config_network(self, custom_networks_config_dict, project, network_choice):
        data = copy.deepcopy(custom_networks_config_dict)

        # Was a bug where couldn't have this name.
        data["networks"]["custom"][0]["name"] = "custom"

        _get_networks_sequence_from_cache.cache_clear()

        with project.temp_config(**data):
            actual = network_choice.convert("ethereum:custom", None, None)

        assert actual.network.name == "custom"

    def test_custom_local_network(self, network_choice):
        uri = "https://example.com"
        actual = network_choice.convert(f"ethereum:local:{uri}", None, None)
        assert actual.uri == uri
        assert actual.network.name == "local"

    def test_explicit_none(self, network_choice):
        actual = network_choice.convert("None", None, None)
        assert actual == _NONE_NETWORK

    def test_bad_ecosystem(self, network_choice):
        # NOTE: "ethereum" is spelled wrong.
        expected = r"No ecosystem named 'etheruem'\. Did you mean 'ethereum'\?"
        with pytest.raises(BadParameter, match=expected):
            network_choice.convert("etheruem:local:test", None, None)

    def test_bad_network(self, network_choice):
        # NOTE: "local" is spelled wrong.
        expected = r"No network in 'ethereum' named 'lokal'\. Did you mean 'local'\?"
        with pytest.raises(BadParameter, match=expected):
            network_choice.convert("ethereum:lokal:test", None, None)

    def test_bad_provider(self, network_choice):
        # NOTE: "test" is spelled wrong.
        expected = (
            r"No provider named 'teest' in network 'local' in "
            r"ecosystem 'ethereum'\. Did you mean 'test'\?"
        )
        with pytest.raises(BadParameter, match=expected):
            network_choice.convert("ethereum:local:teest", None, None)


def test_config_override_option(runner):
    @click.command()
    @config_override_option()
    def cli(config_override):
        assert isinstance(config_override, dict)
        assert config_override["foo"] == "bar"

    result = runner.invoke(cli, ("--config-override", '{"foo": "bar"}'))
    assert result.exit_code == 0
    assert not result.exception



import shutil
from pathlib import Path

import pytest

from tests.conftest import ApeSubprocessRunner, skip_if_plugin_installed
from tests.integration.cli.utils import github_xfail, run_once, skip_projects_except

EXPECTED_FAIL_MESSAGE = "Unknown package '{}'."


@pytest.fixture
def pm_runner(config):
    class PMSubprocessRunner(ApeSubprocessRunner):
        def __init__(self):
            super().__init__("pm", data_folder=config.DATA_FOLDER)

    return PMSubprocessRunner()


@run_once
def test_install_unknown_package(pm_runner):
    path = "path/to/nowhere"
    result = pm_runner.invoke("install", path)
    assert result.exit_code != 0, result.output
    assert EXPECTED_FAIL_MESSAGE.format(path) in result._completed_process.stderr


@run_once
def test_install_path_to_local_package(pm_runner, integ_project):
    project_name = "with-contracts"
    path = Path(__file__).parent / "projects" / project_name
    name = path.stem
    result = pm_runner.invoke("install", path.as_posix(), "--name", project_name)
    assert result.exit_code == 0, result.output
    assert f"Package '{path.as_posix()}' installed."

    # Ensure was installed correctly.
    assert integ_project.dependencies[name]["local"]


@run_once
def test_install_path_to_local_config_file(pm_runner):
    project = "with-contracts"
    path = Path(__file__).parent / "projects" / project / "pyproject.toml"
    arguments = ("install", path.as_posix(), "--name", project)
    result = pm_runner.invoke(*arguments)
    assert result.exit_code == 0, result.output
    assert "SUCCESS" in result.output
    assert "Package 'with-contracts@local' installed." in result.output


@skip_projects_except("test", "with-contracts")
def test_install_local_project_dependencies(pm_runner):
    result = pm_runner.invoke("install")
    assert result.exit_code == 0
    assert "All project packages installed." in result.output


@run_once
def test_install_force(pm_runner):
    result = pm_runner.invoke("install", "--force")
    assert result.exit_code == 0
    assert "All project packages installed." in result.output


@run_once
@github_xfail()
def test_install_github_dependency_with_version(pm_runner):
    result = pm_runner.invoke(
        "install",
        "gh:OpenZeppelin/openzeppelin-contracts",
        "--name",
        "openzeppelin",
        "--version",
        "4.6.0",
        timeout=300,
    )
    assert result.exit_code == 0, result.output
    assert "Package 'openzeppelin@4.6.0' installed."


@run_once
@github_xfail()
def test_install_github_dependency_with_ref(pm_runner):
    result = pm_runner.invoke(
        "install",
        "gh:OpenZeppelin/openzeppelin-contracts",
        "--name",
        "OpenZeppelin",
        "--ref",
        "master",
        timeout=300,
    )
    assert result.exit_code == 0, result.output
    assert "Package 'OpenZeppelin@master' installed."


@skip_projects_except("with-contracts")
def test_install_config_override(pm_runner, integ_project):
    actual_dep = Path(__file__).parent / "projects" / "with-contracts" / "dep"
    shutil.copytree(actual_dep, integ_project.path / "dep")
    config_override = '{"contracts_folder": "src"}'
    dep_path = integ_project.path / "dep"
    name = "foodep2"
    pm_runner.invoke(
        "install",
        dep_path.as_posix(),
        "--name",
        name,
        "--config-override",
        config_override,
        "--force",
    )
    actual = integ_project.dependencies["foodep2"]["local"].config.contracts_folder
    assert actual == "src"


@skip_projects_except("with-contracts")
def test_install_package_id(pm_runner, integ_project):
    pm_runner.project = integ_project
    package_id = [x for x in integ_project.dependencies][0].package_id
    result = pm_runner.invoke("install", package_id)
    assert result.exit_code == 0, result.output


@skip_projects_except("with-contracts")
def test_install_name(pm_runner, integ_project):
    pm_runner.project = integ_project
    name = [x for x in integ_project.dependencies][0].name
    result = pm_runner.invoke("install", name)
    assert result.exit_code == 0, result.output


@skip_projects_except("with-contracts")
def test_install_name_with_version_in_id(pm_runner, integ_project):
    pm_runner.project = integ_project
    dependency = [x for x in integ_project.dependencies][0]
    result = pm_runner.invoke("install", f"{dependency.name}@{dependency.version}")
    assert result.exit_code == 0, result.output
    assert f"Package '{dependency.name}@{dependency.version}' installed." in result.output


@skip_projects_except("with-contracts")
def test_install_name_with_version_flag(pm_runner, integ_project):
    pm_runner.project = integ_project
    dependency = [x for x in integ_project.dependencies][0]
    result = pm_runner.invoke("install", f"{dependency.name}", "--version", dependency.version)
    assert result.exit_code == 0, result.output
    assert f"Package '{dependency.name}@{dependency.version}' installed." in result.output


@skip_projects_except("with-dependencies")
def test_install_no_recurse(pm_runner, integ_project):
    pm_runner.project = integ_project
    result = pm_runner.invoke("install", "--no-recurse")
    assert result.exit_code == 0, result.output
    assert "containing_sub_dependencies/sub_dependency" not in result.output


@run_once
def test_compile_package_not_exists(pm_runner, integ_project):
    pm_runner.project = integ_project
    name = "NOT_EXISTS"
    result = pm_runner.invoke("compile", name)
    expected = f"Dependency '{name}' unknown. Is it installed?"
    assert result.exit_code != 0, result.output
    assert expected in result.output


@skip_projects_except("with-contracts", "with-dependencies")
def test_compile(pm_runner, integ_project):
    pm_runner.project = integ_project
    result = pm_runner.invoke("compile", "--force")
    output = result.output or str(result._completed_process.stderr)
    assert result.exit_code == 0, output

    if integ_project.path.as_posix().endswith("contracts"):
        assert "Package 'foodep@local' compiled." in output, output
    else:
        # Tests against a bug where we couldn't have hyphens in
        # dependency project contracts.
        assert "contracts/hyphen-DependencyContract.json" in output, output


@skip_projects_except("with-contracts")
def test_compile_config_override(pm_runner, integ_project):
    pm_runner.project = integ_project
    cmd = ("compile", "--force", "--config-override", '{"contracts_folder": "src"}')
    result = pm_runner.invoke(*cmd)
    output = result.output or str(result._completed_process.stderr)
    assert result.exit_code == 0, output


@skip_projects_except("with-contracts")
def test_compile_dependency(pm_runner, integ_project):
    pm_runner.project = integ_project
    name = "foodep"

    # Show staring from a clean slate.
    pm_runner.invoke("uninstall", name, "--yes")

    result = pm_runner.invoke("compile", name, "--force")
    assert result.exit_code == 0, result.output
    assert f"Package '{name}@local' compiled." in result.output

    # Show it can happen more than once. (no --force this time).
    result = pm_runner.invoke("compile", name)
    assert result.exit_code == 0, result.output
    assert f"Package '{name}@local' compiled." in result.output

    # Clean for next tests.
    pm_runner.invoke("uninstall", name, "--yes")


@skip_if_plugin_installed("vyper", "solidity")
@skip_projects_except("with-contracts")
def test_compile_missing_compiler_plugins(pm_runner, integ_project, compilers):
    pm_runner.project = integ_project
    name = "depwithunregisteredcontracts"

    # Stateful clean up (just in case?)
    pm_runner.invoke("uninstall", name, "--yes")

    result = pm_runner.invoke("compile", name, "--force")
    expected = (
        "Compiling dependency produced no contract types. "
        "Try installing 'ape-solidity' or 'ape-vyper'"
    )
    assert expected in result.output

    # Also show it happens when installing _all_.
    result = pm_runner.invoke("compile", ".", "--force")
    pm_runner.invoke("uninstall", name, "--yes")

    assert expected in result.output


@skip_projects_except("only-dependencies")
def test_uninstall(pm_runner, integ_project):
    pm_runner.project = integ_project
    package_name = "dependency-in-project-only"

    # Install packages
    pm_runner.invoke("install", ".", "--force")
    result = pm_runner.invoke("uninstall", package_name, "--yes")
    expected_message = f"Uninstalled '{package_name}=local'."
    assert result.exit_code == 0, result.output or result._completed_process.stderr
    assert expected_message in result.output or result._completed_process.stderr


@skip_projects_except("only-dependencies")
def test_uninstall_by_long_name(pm_runner, integ_project):
    pm_runner.project = integ_project
    package_name = "dependency-in-project-only"
    package_long_name = integ_project.dependencies.get_dependency(package_name, "local").package_id

    # Install packages
    pm_runner.invoke("install", ".", "--force")
    result = pm_runner.invoke("uninstall", package_long_name, "--yes")
    expected_message = f"Uninstalled '{package_name}=local'."
    assert result.exit_code == 0, result.output or result._completed_process.stderr
    assert expected_message in result.output or result._completed_process.stderr


@skip_projects_except("only-dependencies")
def test_uninstall_not_exists(pm_runner, integ_project):
    pm_runner.project = integ_project
    package_name = "_this_does_not_exist_"
    result = pm_runner.invoke("uninstall", package_name, "--yes")
    expected_message = f"Package(s) '{package_name}' not installed."
    assert result.exit_code != 0, result.output or result._completed_process.stderr
    assert "ERROR" in result.output
    assert expected_message in result.output


@skip_projects_except("only-dependencies")
def test_uninstall_specific_version(pm_runner, integ_project):
    pm_runner.project = integ_project
    package_name = "dependency-in-project-only"
    version = "local"

    # Install packages
    pm_runner.invoke("install", ".", "--force")
    result = pm_runner.invoke("uninstall", package_name, version, "--yes")
    expected_message = f"Uninstalled '{package_name}=local'."
    assert result.exit_code == 0, result.output
    assert expected_message in result.output


@skip_projects_except("only-dependencies")
def test_uninstall_all_versions(pm_runner, integ_project):
    pm_runner.project = integ_project
    # Install packages
    pm_runner.invoke("install", ".", "--force")
    package_name = "dependency-in-project-only"
    result = pm_runner.invoke("uninstall", package_name, "--yes")
    expected_message = f"Uninstalled '{package_name}=local'."
    assert result.exit_code == 0, result.output
    assert expected_message in result.output


@skip_projects_except("only-dependencies")
def test_uninstall_invalid_version(pm_runner, integ_project):
    pm_runner.project = integ_project
    package_name = "dependency-in-project-only"

    # Install packages
    pm_runner.invoke("install", ".", "--force")

    # Ensure was installed correctly.
    assert package_name in integ_project.dependencies
    assert integ_project.dependencies[package_name]["local"]

    invalid_version = "0.0.0"
    result = pm_runner.invoke("uninstall", package_name, invalid_version, "--yes")
    expected_message = f"Package(s) '{package_name}={invalid_version}' not installed."
    assert result.exit_code != 0, result.output

    assert "ERROR" in result.output
    assert expected_message in result.output


@skip_projects_except("only-dependencies")
def test_uninstall_cancel(pm_runner, integ_project):
    pm_runner.project = integ_project
    package_name = "dependency-in-project-only"
    version = "local"

    # Install packages
    pm_runner.invoke("install", ".", "--force")

    result = pm_runner.invoke("uninstall", package_name, version, input="n\n")
    assert result.exit_code == 0, result._completed_process.stderr
    expected_message = f"Version '{version}' of package '{package_name}' uninstalled."
    assert expected_message not in result.output


@skip_projects_except("only-dependencies")
def test_list(pm_runner, integ_project):
    pm_runner.project = integ_project
    package_name = "dependency-in-project-only"
    dependency = integ_project.dependencies.get_dependency(package_name, "local")

    # Ensure we are not installed.
    dependency.uninstall()

    result = pm_runner.invoke("list")
    assert result.exit_code == 0, result.output

    # NOTE: Not using f-str here so we can see the spacing.
    expected = """
NAME                                VERSION  INSTALLED  COMPILED
apedependencythatisnotinstalledape  <error>  False      False
dependency-in-project-only          local    False      False
    """.strip()
    assert expected in result.output

    # Install and show it change.
    dependency = integ_project.dependencies.get_dependency(package_name, "local")
    dependency.install()

    expected = """
NAME                                VERSION  INSTALLED  COMPILED
apedependencythatisnotinstalledape  <error>  False      False
dependency-in-project-only          local    True       False
    """.strip()
    result = pm_runner.invoke("list")
    assert result.exit_code == 0, result.output
    assert expected in result.output
